searchNodes=[{"doc":"AyeSQL is a library for using raw SQL. Aye /ʌɪ/ exclamation (archaic dialect) : said to express assent; yes. Overview Inspired by Clojure library Yesql , AyeSQL tries to find a middle ground between strings with raw SQL queries and SQL DSLs by: Keeping SQL in SQL files. Generating Elixir functions for every query. Supporting mandatory and optional named parameters. Allowing query composability with ease. Working out of the box with PostgreSQL using Ecto or Postgrex : Being extended to support other databases using the behaviour AyeSQL.Runner . Small Example Let's say we have a SQL query to retrieve the click count of a certain type of link every day of the last X days. In raw SQL this could be written as: WITH computed_dates AS ( SELECT dates::date AS date FROM generate_series( current_date - $1::interval, current_date - interval '1 day', interval '1 day' ) AS dates ) SELECT dates.date AS day, count(clicks.id) AS count FROM computed_dates AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = $2 GROUP BY dates.date ORDER BY dates.date; The equivalent query in Ecto would be: dates = ~s( SELECT generate_series( current_date - ?::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; ) :: date AS d ) from ( c in &quot;clicks&quot; , right_join : day in fragment ( dates , ^ days ) , on : day . d == fragment ( &quot;date(?)&quot; , c . inserted_at ) , where : c . link_id = ^ link_id group_by : day . d , order_by : day . d , select : %{ day : fragment ( &quot;date(?)&quot; , day . d ) , count : count ( c . id ) } ) Using fragments can get convoluted and difficult to maintain. In AyeSQL, the equivalent would be to create an SQL file with the query e.g. queries.sql : -- name: get_day_interval SELECT datetime::date AS date FROM generate_series( current_date - :days::interval, -- Named parameter :days current_date - interval '1 day', interval '1 day' ); -- name: get_avg_clicks -- docs: Gets average click count. WITH computed_dates AS ( :get_day_interval ) -- Composing with another query SELECT dates.date AS day, count(clicks.id) AS count FROM computed_date AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = :link_id -- Named parameter :link_id GROUP BY dates.date ORDER BY dates.date; In Elixir, we would load all the queries in this file by creating the following module: defmodule Queries do use AyeSQL , repo : MyRepo defqueries ( &quot;queries.sql&quot; ) # File name with relative path to SQL file. end or using the macro defqueries/3 : import AyeSQL , only : [ defqueries : 3 ] defqueries ( Queries , &quot;queries.sql&quot; , repo : MyRepo ) Both approaches will create a module called Queries with all the queries defined in queries.sql . And then we could execute the query as follows: iex&gt; params = [ ...&gt; link_id : 42 , ...&gt; days : % Postgrex.Interval { secs : 864_000 } # 10 days ...&gt; ] iex&gt; Queries . get_avg_clicks ( params ) { :ok , [ %{ day : ... , count : ... } , %{ day : ... , count : ... } , %{ day : ... , count : ... } , ... ] }","ref":"AyeSQL.html","title":"AyeSQL","type":"module"},{"doc":"Uses AyeSQL for loading queries. By default, supports the option runner (see AyeSQL.Runner behaviour). Any other option will be passed to the runner.","ref":"AyeSQL.html#__using__/1","title":"AyeSQL.__using__/1","type":"macro"},{"doc":"Macro to load queries from a file . Let's say we have the file lib/sql/queries.sql with the following contents: -- name: get_user -- docs: Gets user by username SELECT * FROM users WHERE username = :username; Then we can load our queries to Elixir using the macro defqueries/1 : # file: lib/queries.ex defmodule Queries do use AyeSQL , repo : MyRepo defqueries ( &quot;sql/queries.sql&quot; ) end or the macro defqueries/3 : # file: lib/queries.ex import AyeSQL , only : [ defqueries : 3 ] defqueries ( Queries , &quot;sql/queries.ex&quot; , repo : MyRepo ) And finally we can inspect the query: iex(1)&gt; Queries . get_user ( username : &quot;some_user&quot; , run : false ) { :ok , % AyeSQL.Query { statement : &quot;SELECT * FROM user WHERE username = $1&quot; , arguments : [ &quot;some_user&quot; ] } } or run it: iex(1)&gt; Queries . get_user ( username : &quot;some_user&quot; ) { :ok , [ %{ username : ... , ... } ] }","ref":"AyeSQL.html#defqueries/1","title":"AyeSQL.defqueries/1","type":"macro"},{"doc":"Macro to load queries from a file and create a module for them. Same as defqueries/1 , but creates a module e.g for the query file lib/sql/queries.sql we can use this macro as follows: # file: lib/queries.ex import AyeSQL , only : [ defqueries : 3 ] defqueries ( Queries , &quot;sql/queries.sql&quot; , repo : MyRepo ) This will generate the module Queries and it'll contain all the SQL statements included in sql/queries.sql .","ref":"AyeSQL.html#defqueries/3","title":"AyeSQL.defqueries/3","type":"macro"},{"doc":"Evaluates the contents of a string with a query and generates an anonyous function that receives parameters and options.","ref":"AyeSQL.html#eval_query/2","title":"AyeSQL.eval_query/2","type":"function"},{"doc":"This module defines the core functionality for AyeSQL.","ref":"AyeSQL.Core.html","title":"AyeSQL.Core","type":"module"},{"doc":"AyeSQL query options.","ref":"AyeSQL.Core.html#t:options/0","title":"AyeSQL.Core.options/0","type":"type"},{"doc":"AyeSQL query parameter name.","ref":"AyeSQL.Core.html#t:parameter_name/0","title":"AyeSQL.Core.parameter_name/0","type":"type"},{"doc":"AyeSQL query parameters.","ref":"AyeSQL.Core.html#t:parameters/0","title":"AyeSQL.Core.parameters/0","type":"type"},{"doc":"This module defines an AyeSQL error.","ref":"AyeSQL.Error.html","title":"AyeSQL.Error","type":"module"},{"doc":"A query struct.","ref":"AyeSQL.Error.html#__struct__/0","title":"AyeSQL.Error.__struct__/0","type":"function"},{"doc":"Creates a new error given some options .","ref":"AyeSQL.Error.html#new/1","title":"AyeSQL.Error.new/1","type":"function"},{"doc":"Query arguments.","ref":"AyeSQL.Error.html#t:arguments/0","title":"AyeSQL.Error.arguments/0","type":"type"},{"doc":"Query errors.","ref":"AyeSQL.Error.html#t:errors/0","title":"AyeSQL.Error.errors/0","type":"type"},{"doc":"Query statement.","ref":"AyeSQL.Error.html#t:statement/0","title":"AyeSQL.Error.statement/0","type":"type"},{"doc":"An error type.","ref":"AyeSQL.Error.html#t:t/0","title":"AyeSQL.Error.t/0","type":"type"},{"doc":"This module defines an AyeSQL query.","ref":"AyeSQL.Query.html","title":"AyeSQL.Query","type":"module"},{"doc":"A query struct.","ref":"AyeSQL.Query.html#__struct__/0","title":"AyeSQL.Query.__struct__/0","type":"function"},{"doc":"Creates a new query given some options .","ref":"AyeSQL.Query.html#new/1","title":"AyeSQL.Query.new/1","type":"function"},{"doc":"Query arguments.","ref":"AyeSQL.Query.html#t:arguments/0","title":"AyeSQL.Query.arguments/0","type":"type"},{"doc":"Query statement.","ref":"AyeSQL.Query.html#t:statement/0","title":"AyeSQL.Query.statement/0","type":"type"},{"doc":"A query type.","ref":"AyeSQL.Query.html#t:t/0","title":"AyeSQL.Query.t/0","type":"type"},{"doc":"This module defines function for expanding the AST.","ref":"AyeSQL.AST.html","title":"AyeSQL.AST","type":"module"},{"doc":"Function to be applied with some parameters and context.","ref":"AyeSQL.AST.html#t:expand_function/0","title":"AyeSQL.AST.expand_function/0","type":"type"},{"doc":"This module defines an AST context.","ref":"AyeSQL.AST.Context.html","title":"AyeSQL.AST.Context","type":"module"},{"doc":"AST context struct.","ref":"AyeSQL.AST.Context.html#__struct__/0","title":"AyeSQL.AST.Context.__struct__/0","type":"function"},{"doc":"Adds a value to the context index.","ref":"AyeSQL.AST.Context.html#add_index/2","title":"AyeSQL.AST.Context.add_index/2","type":"function"},{"doc":"Context id function.","ref":"AyeSQL.AST.Context.html#id/1","title":"AyeSQL.AST.Context.id/1","type":"function"},{"doc":"Merges two contexts.","ref":"AyeSQL.AST.Context.html#merge/2","title":"AyeSQL.AST.Context.merge/2","type":"function"},{"doc":"Merges a context with an error","ref":"AyeSQL.AST.Context.html#merge_error/2","title":"AyeSQL.AST.Context.merge_error/2","type":"function"},{"doc":"Merges a context with a query .","ref":"AyeSQL.AST.Context.html#merge_query/2","title":"AyeSQL.AST.Context.merge_query/2","type":"function"},{"doc":"Creates a new context given some options .","ref":"AyeSQL.AST.Context.html#new/1","title":"AyeSQL.AST.Context.new/1","type":"function"},{"doc":"Updates context with the error not found for a key .","ref":"AyeSQL.AST.Context.html#not_found/2","title":"AyeSQL.AST.Context.not_found/2","type":"function"},{"doc":"Adds arguments in a context given a new value .","ref":"AyeSQL.AST.Context.html#put_argument/2","title":"AyeSQL.AST.Context.put_argument/2","type":"function"},{"doc":"Adds statement in a context given a new value .","ref":"AyeSQL.AST.Context.html#put_statement/2","title":"AyeSQL.AST.Context.put_statement/2","type":"function"},{"doc":"Puts a new variable value in the context .","ref":"AyeSQL.AST.Context.html#put_variable/2","title":"AyeSQL.AST.Context.put_variable/2","type":"function"},{"doc":"Puts several variable value in the context as an SQL list.","ref":"AyeSQL.AST.Context.html#put_variables/2","title":"AyeSQL.AST.Context.put_variables/2","type":"function"},{"doc":"Transforms a context to a query.","ref":"AyeSQL.AST.Context.html#to_query/1","title":"AyeSQL.AST.Context.to_query/1","type":"function"},{"doc":"Argument list.","ref":"AyeSQL.AST.Context.html#t:arguments/0","title":"AyeSQL.AST.Context.arguments/0","type":"type"},{"doc":"Error.","ref":"AyeSQL.AST.Context.html#t:error/0","title":"AyeSQL.AST.Context.error/0","type":"type"},{"doc":"Error type.","ref":"AyeSQL.AST.Context.html#t:error_type/0","title":"AyeSQL.AST.Context.error_type/0","type":"type"},{"doc":"Current context index.","ref":"AyeSQL.AST.Context.html#t:index/0","title":"AyeSQL.AST.Context.index/0","type":"type"},{"doc":"Accumulated statement.","ref":"AyeSQL.AST.Context.html#t:statement/0","title":"AyeSQL.AST.Context.statement/0","type":"type"},{"doc":"AST context.","ref":"AyeSQL.AST.Context.html#t:t/0","title":"AyeSQL.AST.Context.t/0","type":"type"},{"doc":"This module defines functions to compile AyeSQL language strings.","ref":"AyeSQL.Compiler.html","title":"AyeSQL.Compiler","type":"module"},{"doc":"Compiles the contents of a file or string into valid AyeSQL queries.","ref":"AyeSQL.Compiler.html#compile_queries/2","title":"AyeSQL.Compiler.compile_queries/2","type":"function"},{"doc":"Compiles a single query from the contents of a string.","ref":"AyeSQL.Compiler.html#compile_query/2","title":"AyeSQL.Compiler.compile_query/2","type":"function"},{"doc":"Evaluates the contents of a string to an anonymous function with a query that receives parameters and options.","ref":"AyeSQL.Compiler.html#eval_query/2","title":"AyeSQL.Compiler.eval_query/2","type":"function"},{"doc":"Query documentation.","ref":"AyeSQL.Compiler.html#t:docs/0","title":"AyeSQL.Compiler.docs/0","type":"type"},{"doc":"Query fragment.","ref":"AyeSQL.Compiler.html#t:fragment/0","title":"AyeSQL.Compiler.fragment/0","type":"type"},{"doc":"Query fragments.","ref":"AyeSQL.Compiler.html#t:fragments/0","title":"AyeSQL.Compiler.fragments/0","type":"type"},{"doc":"Query name.","ref":"AyeSQL.Compiler.html#t:name/0","title":"AyeSQL.Compiler.name/0","type":"type"},{"doc":"Query parameter.","ref":"AyeSQL.Compiler.html#t:param/0","title":"AyeSQL.Compiler.param/0","type":"type"},{"doc":"Queries.","ref":"AyeSQL.Compiler.html#t:queries/0","title":"AyeSQL.Compiler.queries/0","type":"type"},{"doc":"Query.","ref":"AyeSQL.Compiler.html#t:query/0","title":"AyeSQL.Compiler.query/0","type":"type"},{"doc":"This module defines the lexer for the AyeSQL language.","ref":"AyeSQL.Lexer.html","title":"AyeSQL.Lexer","type":"module"},{"doc":"Gets tokens from the contents of a string.","ref":"AyeSQL.Lexer.html#tokenize/2","title":"AyeSQL.Lexer.tokenize/2","type":"function"},{"doc":"Column number.","ref":"AyeSQL.Lexer.html#t:column/0","title":"AyeSQL.Lexer.column/0","type":"type"},{"doc":"Line number.","ref":"AyeSQL.Lexer.html#t:line/0","title":"AyeSQL.Lexer.line/0","type":"type"},{"doc":"Location.","ref":"AyeSQL.Lexer.html#t:location/0","title":"AyeSQL.Lexer.location/0","type":"type"},{"doc":"Lexer option.","ref":"AyeSQL.Lexer.html#t:option/0","title":"AyeSQL.Lexer.option/0","type":"type"},{"doc":"Lexer options.","ref":"AyeSQL.Lexer.html#t:options/0","title":"AyeSQL.Lexer.options/0","type":"type"},{"doc":"Original match from the string.","ref":"AyeSQL.Lexer.html#t:original/0","title":"AyeSQL.Lexer.original/0","type":"type"},{"doc":"Token.","ref":"AyeSQL.Lexer.html#t:token/0","title":"AyeSQL.Lexer.token/0","type":"type"},{"doc":"Token name.","ref":"AyeSQL.Lexer.html#t:token_name/0","title":"AyeSQL.Lexer.token_name/0","type":"type"},{"doc":"Tokens.","ref":"AyeSQL.Lexer.html#t:tokens/0","title":"AyeSQL.Lexer.tokens/0","type":"type"},{"doc":"Token value.","ref":"AyeSQL.Lexer.html#t:value/0","title":"AyeSQL.Lexer.value/0","type":"type"},{"doc":"This module defines an AyeSQL.Runner .","ref":"AyeSQL.Runner.html","title":"AyeSQL.Runner","type":"behaviour"},{"doc":"Uses the AyeSQL.Runner behaviour.","ref":"AyeSQL.Runner.html#__using__/1","title":"AyeSQL.Runner.__using__/1","type":"macro"},{"doc":"Callback to initialize the runner.","ref":"AyeSQL.Runner.html#c:run/2","title":"AyeSQL.Runner.run/2","type":"callback"},{"doc":"This module defines Ecto default adapter. Can be used as follows: defmodule MyQueries do use AyeSQL , repo : MyRepo defqueries ( &quot;query/my_queries.sql&quot; ) end","ref":"AyeSQL.Runner.Ecto.html","title":"AyeSQL.Runner.Ecto","type":"module"},{"doc":"This module defines Postgrex default adapter. Can be used as follows: defmodule MyQueries do use AyeSQL , runner : AyeSQL.Runner.Postgrex defqueries ( &quot;query/my_queries.sql&quot; ) end And given a connection to the database, then it can be used with the query options: iex&gt; MyQueries . get_user ( [ id : id ] , conn : connection ) { :ok , ... }","ref":"AyeSQL.Runner.Postgrex.html","title":"AyeSQL.Runner.Postgrex","type":"module"},{"doc":"This module defines an AyeSQL compile error exception.","ref":"AyeSQL.CompileError.html","title":"AyeSQL.CompileError","type":"exception"},{"doc":"A compiler error.","ref":"AyeSQL.CompileError.html#__struct__/0","title":"AyeSQL.CompileError.__struct__/0","type":"function"},{"doc":"A compile error.","ref":"AyeSQL.CompileError.html#t:t/0","title":"AyeSQL.CompileError.t/0","type":"type"},{"doc":"Aye /ʌɪ/ exclamation (archaic dialect) : said to express assent; yes. AyeSQL is a library for using raw SQL.","ref":"readme.html","title":"AyeSQL","type":"extras"},{"doc":"Inspired by Clojure library Yesql , AyeSQL tries to find a middle ground between strings with raw SQL queries and SQL DSLs. This library aims to: Keep SQL in SQL files. Generate easy to use Elixir functions for every query. Parameterize queries using maps and keyword lists. Allow query composablity. Work out-of-the-box with PostgreSQL using Ecto or Postgrex . If you want to know more about AyeSQL: Small example Syntax Naming queries Parameters Mandatory parameters Query composition Optional fragments IN statement Subqueries and subfragments Installation And the following additional links provide more information about the library: Full Documentation AyeSQL: Writing Raw SQL in Elixir Why raw SQL? Dynamic queries with EEx Adding support to other databases","ref":"readme.html#overview","title":"AyeSQL - Overview","type":"extras"},{"doc":"In AyeSQL, the equivalent would be to create an SQL file with the query e.g. queries.sql : -- file: queries.sql -- name: get_avg_clicks -- docs: Gets average click count. WITH computed_dates AS ( SELECT datetime::date AS date FROM generate_series( current_date - :days::interval, -- Named parameter :days current_date - interval '1 day', interval '1 day' ) ) SELECT dates.date AS day, count(clicks.id) AS count FROM computed_date AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = :link_id -- Named parameter :link_id GROUP BY dates.date ORDER BY dates.date; In Elixir, we would load all the queries in this file by creating the following module: # file: lib/queries.ex defmodule Queries do use AyeSQL , repo : MyRepo defqueries ( &quot;queries.sql&quot; ) # File name with relative path to SQL file. end or using the macro defqueries/3 : # file: lib/queries.ex import AyeSQL , only : [ defqueries : 3 ] defqueries ( Queries , &quot;queries.sql&quot; , repo : MyRepo ) Note : The file name used in defqueries macro should be relative to the file where the macro is used. Both approaches will create a module called Queries with all the queries defined in queries.sql . And then we could execute the query as follows: iex&gt; params = [ ...&gt; link_id : 42 , ...&gt; days : % Postgrex.Interval { secs : 864_000 } # 10 days ...&gt; ] iex&gt; Queries . get_avg_clicks ( params ) { :ok , [ %{ day : ... , count : ... } , %{ day : ... , count : ... } , %{ day : ... , count : ... } , ... ] }","ref":"readme.html#small-example","title":"AyeSQL - Small Example","type":"extras"},{"doc":"An SQL file can have as many queries as you want as long as they are named. For the following sections we'll assume we have: lib/my_repo.ex which is an Ecto repo called MyRepo . lib/queries.sql with SQL queries. lib/queries.ex with the following structure: import AyeSQL , only : [ defqueries : 3 ] defqueries ( Queries , &quot;queries.sql&quot; , repo : MyRepo ) Naming Queries For naming queries, we add a comment with the keyword -- name: followed by the name of the function e.g the following query would generate the function Queries.get_hostnames/2 : -- name: get_hostnames SELECT hostname FROM server Additionally, we could also add documentation for the query by adding a comment with the keyword -- docs: followed by the query's documentation e.g: -- name: get_hostnames -- docs: Gets hostnames from the servers. SELECT hostname FROM server Important: if the function does not have -- docs: it won't have documentation e.g. @doc false . Parameters There are two types of parameters: Mandatory: for passing parameters to a query. They start with : e.g. :hostname . Optional: for query composability. They start with :_ e.g. :_order_by . Additionally, any query in a file can be accessed with its name adding : at the front e.g :get_hostnames . Mandatory Parameters Let's say we want to get the name of an operative system by architecture: -- name: get_os_by_architecture -- docs: Gets operative system's name by a given architecture. SELECT name FROM operative_system WHERE architecture = :architecture The previous query would generate the function Queries.get_os_by_architecture/2 that can be called as: iex&gt; Queries . get_os_by_architecture ( architecture : &quot;AMD64&quot; ) { :ok , [ %{ name : &quot;Debian Buster&quot; } , %{ name : &quot;Windows 10&quot; } , ... ] } Query Composition Now if we would like to get hostnames by architecture we could compose queries by doing the following: -- name: get_os_by_architecture -- docs: Gets operative system's name by a given architecture. SELECT name FROM operative_system WHERE architecture = :architecture -- name: get_hostnames_by_architecture -- docs: Gets hostnames by architecture. SELECT hostname FROM servers WHERE os_name IN ( :get_os_by_architecture ) The previous query would generate the function Queries.get_hostnames_by_architecture/2 that can be called as: iex&gt; Queries . get_hostnames_by_architecture ( architecture : &quot;AMD64&quot; ) { :ok , [ %{ hostname : &quot;server0&quot; } , %{ hostname : &quot;server1&quot; } , ... ] } Optional Fragments Let's say that now we need to order ascending or descending by hostname by using an optional :_order_by parameter e.g: -- name: get_os_by_architecture -- docs: Gets operative system's name by a given architecture. SELECT name FROM operative_system WHERE architecture = :architecture -- name: get_hostnames_by_architecture -- docs: Gets hostnames by architecture. SELECT hostname FROM servers WHERE os_name IN ( :get_os_by_architecture ) :_order_by -- name: ascending ORDER BY hostname ASC -- name: descending ORDER BY hostname DESC The previous query could be called as before: iex&gt; Queries . get_hostnames_by_architecture ( architecture : &quot;AMD64&quot; ) { :ok , [ %{ hostname : &quot;Barcelona&quot; } , %{ hostname : &quot;Granada&quot; } , %{ hostname : &quot;Madrid&quot; } , ... ] } or by order ascending: iex&gt; params = [ architecture : &quot;AMD64&quot; , _order_by : :ascending ] iex&gt; Queries . get_hostnames_by_architecture ( params ) { :ok , [ %{ hostname : &quot;Barcelona&quot; } , %{ hostname : &quot;Madrid&quot; } , %{ hostname : &quot;Granada&quot; } , ... ] } or descending: iex&gt; params = [ architecture : &quot;AMD64&quot; , _order_by : :descending ] iex&gt; Queries . get_hostnames_by_architecture ( params ) { :ok , [ %{ hostname : &quot;Zaragoza&quot; } , %{ hostname : &quot;Madrid&quot; } , %{ hostname : &quot;Granada&quot; } , ... ] } Important: A query can be called by name e.g. :descending if it's defined in the same SQL file. Otherwise, we need to pass the function instead e.g. Queries.descending/2 iex&gt; params = [ architecture : &quot;AMD64&quot; , _order_by : Queries . descending / 2 ] iex&gt; Queries . get_hostnames_by_architecture ( params ) { :ok , [ %{ hostname : &quot;Zaragoza&quot; } , %{ hostname : &quot;Madrid&quot; } , %{ hostname : &quot;Granada&quot; } , ... ] } IN Statement Lists in SQL might be tricky. That's why AyeSQL supports a special type for them e.g: Let's say we have the following query: -- name: get_os_by_hostname -- docs: Gets hostnames and OS names given a list of hostnames. SELECT hostname, os_name FROM servers WHERE hostname IN (:hostnames) It is possible to do the following: iex&gt; params = [ hostnames : { :in , [ &quot;server0&quot; , &quot;server1&quot; , &quot;server2&quot; ] } ] iex&gt; Server . get_os_by_hostname ( params ) { :ok , [ %{ hostname : &quot;server0&quot; , os_name : &quot;Debian Buster&quot; } , %{ hostname : &quot;server1&quot; , avg_ram : &quot;Windows 10&quot; } , %{ hostname : &quot;server2&quot; , avg_ram : &quot;Minix 3&quot; } ] } Subqueries and Subfragments Subqueries can be composed directly, as show before, or via the :inner tuple e.g. let's say we need to get the adults order by name in ascending order and age in descending order: -- name: ascending ASC -- name: descending DESC -- name: by_age age :order_direction -- name: by_name name :order_direction -- name: get_adults -- docs: Gets adults. SELECT name, age FROM person WHERE age &gt;= 18 ORDER BY :order_by Then our code in elixir would be: iex&gt; order_by = [ ...&gt; by_name : [ order_direction : :ascending ] , ...&gt; by_age : [ order_direction : :descending ] ...&gt; ] iex&gt; Queries . get_adults ( order_by : { :inner , order_by , &quot;, &quot; } ) { :ok , [ %{ name : &quot;Alice&quot; , age : 42 } , %{ name : &quot;Bob&quot; , age : 21 } , ... ] } Note : If you're using this level of composability and it fits your use case, consider using either: Ecto EEx templates","ref":"readme.html#syntax","title":"AyeSQL - Syntax","type":"extras"},{"doc":"AyeSQL is available as a Hex package. To install, add it to your dependencies in your mix.exs file: def deps do [ { :ayesql , &quot;~&gt; 1.1&quot; } ] end If you're going to use any of the provided query runners, then you should add their dependencies as well: Add :ecto_sql for AyeSQL.Runner.Ecto (default runner). Add :postgrex for AyeSQL.Runner.Postgrex . Add :ecto_sql and :postgrex for running queries using Ecto in a PostgreSQL database.","ref":"readme.html#installation","title":"AyeSQL - Installation","type":"extras"},{"doc":"Alexander de Sousa.","ref":"readme.html#author","title":"AyeSQL - Author","type":"extras"},{"doc":"AyeSQL is released under the MIT License. See the LICENSE file for further details.","ref":"readme.html#license","title":"AyeSQL - License","type":"extras"},{"doc":"Writing and running raw SQL in Elixir is not pretty. Not only the lack of syntax highlighting is horrible, but also substituting parameters into the query string can be unmaintainable e.g: query = &quot;&quot;&quot; SELECT hostname, AVG(ram_usage) AS avg_ram FROM server WHERE hostname IN ($1, $2, $3) AND location = $4 GROUP BY hostname &quot;&quot;&quot; arguments = [ &quot;server_0&quot; , &quot;server_1&quot; , &quot;server_2&quot; , &quot;Barcelona&quot; ] Postgrex . query ( conn , query , arguments ) Adding more hostnames to the previous query would be a nightmare. If the arguments are generated dynamically, then editing this query would be a challenging task. Thankfully, we have Ecto , which provides a great DSL for generating database queries at runtime. The same query in Ecto would be the following: servers = [ &quot;server_0&quot; , &quot;server_1&quot; , &quot;server_2&quot; ] location = &quot;Barcelona&quot; from s in &quot;server&quot; , where : s . location == ^ location and s . hostname in ^ servers , select : %{ hostname : s . hostname , avg_ram : avg ( s . ram_usage ) } Pretty straightforward and maintainable. If Ecto is so good for building queries, why would you use raw SQL? . Though Ecto is quite good with simple queries, complex custom queries often require the use of fragments. Fragments are not pretty though there are workarounds using macros to make them prettier. It's easier to see with an example: let's say we have to retrieve the click count of a certain type of link every day of the last N days . With complex queries, developers tend to start writing them in raw SQL: WITH computed_dates AS ( SELECT dates::date AS date FROM generate_series( current_date - $1::interval, current_date - interval '1 day', interval '1 day' ) AS dates ) SELECT dates.date AS day, count(clicks.id) AS count FROM computed_dates AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = $2 GROUP BY dates.date ORDER BY dates.date; Once we have the raw SQL, it's a bit easier to write our Ecto query. In this case, this query should be written using fragments: dates = &quot;&quot;&quot; SELECT generate_series( current_date - ?::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; )::date AS d &quot;&quot;&quot; from ( c in &quot;clicks&quot; , right_join : day in fragment ( dates , ^ days ) , on : day . d == fragment ( &quot;date(?)&quot; , c . inserted_at ) , where : c . link_id = ^ link_id group_by : day . d , order_by : day . d , select : %{ day : fragment ( &quot;date(?)&quot; , day . d ) , count : count ( c . id ) } ) It's not the ideal solution (yet) and it's harder to maintain than the raw SQL solution that would work out-of-the-box. Only to get to the previous solution, the developer need to: Know the specific SQL dialect of the database they're using. Know Ecto's API and its limitations. For both, raw SQL and Ecto query, the end result for this query would be the same. With extra effort we found a subpar solution that gives us the same result as our raw SQL. The final, and sometimes, optional step would be to transform the Ecto query into something a bit more maintainable. defmodule CustomDSL do defmacro date ( date ) do quote do fragment ( &quot;date(?)&quot; , unquote ( date ) ) end end defmacro ndays ( n ) do query = &quot;&quot;&quot; SELECT generate_series( current_date - ?::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; )::date AS d &quot;&quot;&quot; quote do fragment ( unquote ( query ) , unquote ( n ) ) end end end import CustomDSL from ( c in &quot;clicks&quot; , right_join : day in ndays ( ^ days ) , on : day . d == date ( c . inserted_at ) , where : c . link_id = ^ link_id group_by : day . d , order_by : day . d , select : %{ day : date ( day . d ) count : count ( c . id ) } ) The previous query is more readable, but requires knowledge of: The specific SQL dialect. Ecto's API and its limitations. Elixir's macros. Custom DSL API. For some problems, getting to this final stage is preferable. However, for some other problems, the raw SQL query would have been enough. The raw SQL query was already a good solution to the problem. It only needs a maintainable way to be parameterized. That's why AyeSQL exists.","ref":"why-raw-sql.html","title":"Why Raw SQL","type":"extras"},{"doc":"Queries are run by default. To avoid this and get the AyeSQL.Query.t() struct instead, set the option run: false on the top of the module e.g: defmodule Queries do use AyeSQL , repo : MyRepo , run : false defqueries ( &quot;myqueries.sql&quot; ) end or when running a query: iex&gt; params = [ ...&gt; link_id : 42 , ...&gt; days : % Postgrex.Interval { secs : 864_000 } # 10 days ...&gt; ] iex&gt; Queries . get_avg_clicks ( params , run : false ) { :ok , % AyeSQL.Query { statement : ... , arguments : ... } }","ref":"debugging-queries.html","title":"Debugging Queries","type":"extras"},{"doc":"When AyeSQL syntax for dynamic query generation falls short, we can leverage EEx templates for generating them. Note: It's not possible to use defqueries/1 and defqueries/3 macros when using EEx templates. We'll need to use the function AyeSQL.eval_query/2 instead. This function will generate an anonymous function with the query. One of such cases is when the identifiers in the query are not constant (table names, aliases, etc). The following EEx template has a query with dynamic tables, aliases, fragments and even joins: -- File: queries.sql.eex &lt;% [%{name: name, as: as} | rest] = @tables %&gt; SELECT &lt;%= as %&gt;.datetime AS datetime, &lt;%= @calculation %&gt; AS calculation FROM &lt;%= name %&gt; AS &lt;%= as %&gt; &lt;%= for table &lt;- rest do %&gt; INNER JOIN &lt;%= table.name %&gt; AS &lt;%= table.as %&gt; ON &lt;%= table.as %&gt;.datetime = &lt;%= as %&gt;.datetime &lt;% end %&gt; WHERE &lt;%= as %&gt;.datetime BETWEEN :start_date AND :end_date With the previous template, we can generate a valid AyeSQL query with the following: iex&gt; assigns = [ ...&gt; calculation : &quot;(a.value / b.value)&quot; , ...&gt; tables : [ ...&gt; %{ name : &quot;table_a&quot; , as : &quot;a&quot; } , ...&gt; %{ name : &quot;table_b&quot; , as : &quot;b&quot; } ...&gt; ] ...&gt; ] iex&gt; dynamic_query = ...&gt; &quot;queries.sql.eex&quot; ...&gt; |&gt; File . read! ( ) ...&gt; |&gt; EEx . eval_string ( assigns : assigns ) so, the variable dynamic_query contains the following query: SELECT a.datetime AS datetime, (a.value / b.value) AS calculation FROM table_a AS a INNER JOIN table_b AS b ON b.datetime = a.datetime WHERE a.datetime BETWEEN :start_date AND :end_date This query is then compatible with AyeSQL , so we can compile it into a function using the following: iex&gt; function = AyeSQL . eval_query ( dynamic_query ) And finally use the function as any other AyeSQL function: iex&gt; { :ok , start_date , _ } = DateTime . from_iso8601 ( &quot;2022-04-17T00:00:00Z&quot; ) iex&gt; function . ( [ start_date : start_date , end_date : DateTime . utc_now ( ) ] , repo : MyApp.Repo ) { :ok , [ %{ datetime : ... , calculation : ... } , ... ] }","ref":"dynamic-queries-with-eex.html","title":"Dynamic Queries With EEx","type":"extras"},{"doc":"AyeSQL can be extended to support more databases than PostgreSQL by implementing the behaviour AyeSQL.Runner . This library ships with two runners already: AyeSQL.Runner.Ecto , (it's the default and assumes PostgreSQL is the database). AyeSQL.Runner.Postgrex . By default, AyeSQL uses the runner AyeSQL.Runner.Ecto . This runner only has one option which is :repo for the repo module. Additionally, it converts the result to a list of maps. In order to support your database, you just need to implement the behaviour e.g. let's say we implement a MySQL our queries module would be created using something like the following: defmodule MyApp.Queries do use AyeSQL , runner : MyApp.Runner.MySQL , pool : MyApp.MySQL.Pool defqueries ( &quot;queries.sql&quot; ) end or import AyeSQL , only : [ defqueries : 3 ] defqueries ( Queries , &quot;queries.sql&quot; , runner : MyApp.Runner.MySQL , pool : MyApp.MySQL.Pool ) And our runner could be implemented like this: defmodule MyApp.Runner.MySQL do use AyeSQL.Runner @impl AyeSQL def run ( % AyeSQL.Query { statement : stmt , arguments : args } , options ) do query_options = Keyword . frop ( options , [ :pool , :into ] ) stmt = transform_stmt ( stmt ) with { :ok , result } &lt;- MyXQL . query ( pool , stmt , args , query_options ) do result = AyeSQL.Runner . handle_result ( result ) { :ok , result } end end @spec transform_stmt ( AyeSQL.Query . statement ( ) ) :: AyeSQL.Query . statement ( ) defp transform_stmt ( stmt ) do Regex . replace ( ~r/ \\$ ( \\d +)/ , stmt , &quot;?&quot; ) end end","ref":"query-runners.html","title":"Query Runners","type":"extras"},{"doc":"","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Enhancements Added new lexer with improved tokens. Now these tokens track: lines and columns for each token. original and modified contents. Added new parser with support for both named and unamed queries. Improved compiler errors. Now they show: line and columns where errors occur. surrounding lines containing the error. Added missing tests for the runners. Reorganized documentation.","ref":"changelog.html#v-1-1-0","title":"Changelog - v 1.1.0","type":"extras"},{"doc":"Enhancements Added support for specific runner options e.g. adding a timeout to the Postgrex query. Breaking changes Now it runs queries by default. To avoid running the queries, add the option run: false on the module or the function calls.","ref":"changelog.html#v1-0-0","title":"Changelog - v1.0.0","type":"extras"},{"doc":"Bug fix Dax Ravi fixed an infinite recursion problem in fragment token regular expression in the lexer.","ref":"changelog.html#v0-6-1","title":"Changelog - v0.6.1","type":"extras"},{"doc":"Enhancements Added support for subqueries with local scope. Improved documentation.","ref":"changelog.html#v0-6-0","title":"Changelog - v0.6.0","type":"extras"},{"doc":"Bug fix Added missing support for function calls by name in parameters.","ref":"changelog.html#v0-5-5","title":"Changelog - v0.5.5","type":"extras"},{"doc":"Bug fix Added support for nil values for parameters ( NULL ).","ref":"changelog.html#v0-5-4","title":"Changelog - v0.5.4","type":"extras"},{"doc":"Bug fix Added support for Windows new line.","ref":"changelog.html#v0-5-3","title":"Changelog - v0.5.3","type":"extras"},{"doc":"Enhancements Fixed dialyzer warnings for unexistent type AyeSQL.Core.query() .","ref":"changelog.html#v0-5-2","title":"Changelog - v0.5.2","type":"extras"},{"doc":"Enhancements Improved documentation.","ref":"changelog.html#v0-5-1","title":"Changelog - v0.5.1","type":"extras"},{"doc":"This version maintains the same query language, but it breaks runners as now queries return AyeSQL.Query.t() instead of {binary(), [term()]} . Enhancements Refactored code to improve readability. Added AyeSQL.Query for queries (instead of tuple). Added AyeSQL.Error for query errors (instead of returning a string with the missing parameter). Added AyeSQL.AST.Context to be able to handle several errors, instead of returning them one at the time. Improved documentation explaining all features.","ref":"changelog.html#v0-5-0","title":"Changelog - v0.5.0","type":"extras"},{"doc":"Enhancements Added optional arguments for better composition.","ref":"changelog.html#v0-4-1","title":"Changelog - v0.4.1","type":"extras"},{"doc":"Enhancements Added AyeSQL.Runner behaviour for writing custom query runners. New query runners for Ecto and Postgrex connections.","ref":"changelog.html#v0-4-0","title":"Changelog - v0.4.0","type":"extras"},{"doc":"Added configuration support for running queries by default.","ref":"changelog.html#v0-3-2","title":"Changelog - v0.3.2","type":"extras"},{"doc":"Enhancements Added defqueries/3 to avoid boiler plate code.","ref":"changelog.html#v0-3-1","title":"Changelog - v0.3.1","type":"extras"},{"doc":"Enhancements Simplified lexer and parser. Now it's possible to accept anonymous blocks of code.","ref":"changelog.html#v0-3-0","title":"Changelog - v0.3.0","type":"extras"},{"doc":"Enhancements Now it detects when an SQL file has been changed (suggested by Ole Morten Halvorsen ). Updated dependencies. Improved code for maintainability. Improved tests for more code coverage. Support for Elixir 1.8 and Erlang 21","ref":"changelog.html#v0-2-0","title":"Changelog - v0.2.0","type":"extras"}]